from .lxd import convert_lxd

from pathlib import Path
from typing import Annotated

import typer

__all__ = ["app", "convert_lxd"]

app = typer.Typer(pretty_exceptions_show_locals=False)


@app.command("from-lxd", help="Convert datasets from LevelXData to omega-prime.")
def convert_lxd_cli(
    dataset_path: Annotated[
        Path,
        typer.Argument(
            exists=True, dir_okay=True, file_okay=False, readable=True, help="Root of the LevelXData dataset"
        ),
    ],
    output_path: Annotated[
        Path,
        typer.Argument(file_okay=False, writable=True, help="In which folder to write the created omega-prime files"),
    ],
    n_workers: Annotated[int, typer.Option(help="Set to -1 for n_cpus-1 workers.")] = 1,
):
    Path(output_path).mkdir(exist_ok=True)
    convert_lxd(dataset_dir=dataset_path, outpath=output_path, n_workers=n_workers)



import typing
from pathlib import Path
from warnings import warn

import betterosi
import numpy as np
import shapely
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon as PltPolygon
import pandas as pd
import polars as pl
import altair as alt

import pandera as pa
import pandera.extensions as extensions
import pyarrow
import pyarrow.parquet as pq
import polars as pl
from .map_odr import MapOdr
from .map import MapOsi, ProjectionOffset, MapOsiCenterline
import itertools
from functools import partial
import altair as alt
import polars as pl
import polars_st as st

pi_valued = pa.Check.between(-np.pi, np.pi)


@extensions.register_check_method(
    statistics=["column_name", "column_value", "other_column_name", "other_column_unset_value"]
)
def other_column_set_on_column_value(
    pandas_obj, *, column_name: str, column_value, other_column_name: str, other_column_unset_value
):
    return ~np.logical_and(
        pandas_obj[column_name] == column_value, pandas_obj[other_column_name] == other_column_unset_value
    )


@extensions.register_check_method(
    statistics=["column_name", "column_value", "other_column_name", "other_column_unset_value"]
)
def other_column_unset_on_column_value(
    pandas_obj, *, column_name, column_value, other_column_name: str, other_column_unset_value: int
):
    return ~np.logical_and(
        pandas_obj[column_name] != column_value, pandas_obj[other_column_name] != other_column_unset_value
    )


recording_moving_object_schema = pa.DataFrameSchema(
    title="DataFrame Schema for ASAM OSI GroundTruth of MovingObjects",
    description="",
    columns={
        "x": pa.Column(float, title="MovingObject.base.position.x", description="osi3.Vector3d.x"),
        "y": pa.Column(float, title="MovingObject.base.position.y", description="osi3.Vector3d.y"),
        "z": pa.Column(float, title="MovingObject.base.position.z", description="osi3.Vector3d.z"),
        "vel_x": pa.Column(float, title="MovingObject.base.velocity.x", description="osi3.Vector3d.x"),
        "vel_y": pa.Column(float, title="MovingObject.base.velocity.y", description="osi3.Vector3d.y"),
        "vel_z": pa.Column(float, title="MovingObject.base.velocity.z", description="osi3.Vector3d.z"),
        "acc_x": pa.Column(float, title="MovingObject.base.acceleration.x", description="osi3.Vector3d.x"),
        "acc_y": pa.Column(float, title="MovingObject.base.acceleration.y", description="osi3.Vector3d.y"),
        "acc_z": pa.Column(float, title="MovingObject.base.acceleration.z", description="osi3.Vector3d.z"),
        "length": pa.Column(
            float, pa.Check.gt(0), title="MovingObject.base.dimesion.length", description="osi3.Dimenstion3d.length"
        ),
        "width": pa.Column(
            float, pa.Check.gt(0), title="MovingObject.base.dimesion.width", description="osi3.Dimenstion3d.width"
        ),
        "height": pa.Column(
            float, pa.Check.ge(0), title="MovingObject.base.dimesion.height", description="osi3.Dimenstion3d.height"
        ),
        "type": pa.Column(
            int,
            pa.Check.between(
                0, 4, error=f"Type must be one of { ({o.name: o.value for o in betterosi.MovingObjectType}) }"
            ),
            title="MovingObject.type",
            description="osi3.MovingObject.Type",
        ),
        "role": pa.Column(
            int,
            pa.Check.between(
                -1,
                10,
                error=f"Type must be one of { ({o.name: o.value for o in betterosi.MovingObjectVehicleClassificationRole}) }",
            ),
            title="MovingObject.vehicle_classification.role",
            description="osi3.MovingObject.VehicleClassification.Role",
        ),
        "subtype": pa.Column(
            int,
            pa.Check.between(
                -1,
                17,
                error=f"Subtype must be one of { ({o.name: o.value for o in betterosi.MovingObjectVehicleClassificationType}) }",
            ),
            title="MovingObject.vehicle_classification.type",
            description="osi3.MovingObject.VehicleClassification.Type",
        ),
        "roll": pa.Column(
            float, pi_valued, title="MovingObject.base.orientation.roll", description="osi3.Orientation3d.roll"
        ),
        "pitch": pa.Column(
            float, pi_valued, title="MovingObject.base.orientation.pitch", description="osi3.Orientation3d.pitch"
        ),
        "yaw": pa.Column(
            float, pi_valued, title="MovingObject.base.orientation.yaw", description="osi3.Orientation3d.yaw"
        ),
        "idx": pa.Column(int, pa.Check.ge(0), title="MovingObject.id.value", description="osi3.Identifier.value"),
        "total_nanos": pa.Column(
            int,
            pa.Check.ge(0),
            title="GroundTruth.timestamp.nanos+1e9*GroundTruth.timestamp.seconds",
            description="osi3.Timestamp.nanos, osi3.Timestamp.seconds",
        ),
    },
    unique=["idx", "total_nanos"],
    checks=[
        pa.Check.other_column_set_on_column_value(
            "type",
            int(betterosi.MovingObjectType.TYPE_VEHICLE),
            "role",
            -1,
            error="`role` is `-1` despite type beeing `TYPE_VEHICLE`",
        ),
        pa.Check.other_column_unset_on_column_value(
            "type",
            int(betterosi.MovingObjectType.TYPE_VEHICLE),
            "role",
            -1,
            error="`role` is set despite type not beeing `TYPE_VEHICLE`",
        ),
        pa.Check.other_column_set_on_column_value(
            "type",
            int(betterosi.MovingObjectType.TYPE_VEHICLE),
            "subtype",
            -1,
            error="`subtype` is `-1` despite type beeing `TYPE_VEHICLE`",
        ),
        pa.Check.other_column_unset_on_column_value(
            "type",
            int(betterosi.MovingObjectType.TYPE_VEHICLE),
            "subtype",
            -1,
            error="`subtype` is set despite type not beeing `TYPE_VEHICLE`",
        ),
    ],
)


def timestamp2ts(timestamp: betterosi.Timestamp):
    return timestamp.seconds * 1_000_000_000 + timestamp.nanos


def nearest_interp(xi, x, y):
    # https://stackoverflow.com/a/21003629
    idx = np.abs(x - xi[:, None])
    return y[idx.argmin(axis=1)]


class MovingObject:
    def __init__(self, recording, idx):
        super().__init__()
        self.idx = int(idx)
        self._recording = recording
        self._df = self._recording._df.filter(pl.col("idx") == self.idx)

        for k in [
            "x",
            "y",
            "z",
            "vel_x",
            "vel_y",
            "vel_z",
            "acc_x",
            "acc_y",
            "acc_z",
            "yaw",
            "roll",
            "pitch",
            "vel",
            "acc",
        ]:
            setattr(self, k, self._df[k])
        self.timestamps = self._df["total_nanos"] / 1e9
        for k in ["length", "width", "height"]:
            setattr(self, f"{k}s", self._df[k])
            setattr(self, k, self._df[k].mean())

        self.type = betterosi.MovingObjectType(self._df["type"][0])
        subtype_int = self._df["subtype"][0]
        self.subtype = betterosi.MovingObjectVehicleClassificationType(subtype_int) if subtype_int != -1 else None
        role_int = self._df["role"][0]
        self.role = betterosi.MovingObjectVehicleClassificationRole(role_int) if role_int != -1 else None
        self.birth = int(self._df["frame"][0])
        self.end = int(self._df["frame"][-1])

    # def set(self, k, val):
    #    self._recording._df.loc[self._recording._df["idx"] == self.idx, k] = val

    @property
    def polygon(self):
        if "polygon" not in self._df.columns:
            self._recording._add_polygons_to_df()
            self._df = self._recording._df.filter(pl.col("idx") == self.idx)
        return self._df["polygon"]

    @property
    def polygons(self):
        return self.polygon

    @property
    def nanos(self):
        return self.timestamps * 1e9

    def plot(self, ax: plt.Axes):
        ax.plot(self.x, self.y, label=str(self.idx), c="red", alpha=0.5)
        pass

    def plot_mv_frame(self, ax: plt.Axes, frame: int):
        if "polygon" not in self._df.columns:
            self._recording._add_polygons_to_df()
        polys = self._df.filter(pl.col("frame") == frame)["polygon"]
        for p in polys:
            ax.add_patch(PltPolygon(p.exterior.coords, fc="red", alpha=0.2))


class Recording:
    _MovingObjectClass: typing.ClassVar = MovingObject

    @staticmethod
    def _add_polygons(df):
        if "polygon" not in df.columns:
            ar = (
                df[:]
                .select(
                    (
                        pl.col("x")
                        + (+pl.col("length") / 2) * pl.col("yaw").cos()
                        - (+pl.col("width") / 2) * pl.col("yaw").sin()
                    ).alias("x1"),
                    (
                        pl.col("x")
                        + (+pl.col("length") / 2) * pl.col("yaw").cos()
                        - (-pl.col("width") / 2) * pl.col("yaw").sin()
                    ).alias("x2"),
                    (
                        pl.col("x")
                        + (-pl.col("length") / 2) * pl.col("yaw").cos()
                        - (-pl.col("width") / 2) * pl.col("yaw").sin()
                    ).alias("x3"),
                    (
                        pl.col("x")
                        + (-pl.col("length") / 2) * pl.col("yaw").cos()
                        - (+pl.col("width") / 2) * pl.col("yaw").sin()
                    ).alias("x4"),
                    (
                        pl.col("y")
                        + (+pl.col("length") / 2) * pl.col("yaw").sin()
                        + (+pl.col("width") / 2) * pl.col("yaw").cos()
                    ).alias("y1"),
                    (
                        pl.col("y")
                        + (+pl.col("length") / 2) * pl.col("yaw").sin()
                        + (-pl.col("width") / 2) * pl.col("yaw").cos()
                    ).alias("y2"),
                    (
                        pl.col("y")
                        + (-pl.col("length") / 2) * pl.col("yaw").sin()
                        + (-pl.col("width") / 2) * pl.col("yaw").cos()
                    ).alias("y3"),
                    (
                        pl.col("y")
                        + (-pl.col("length") / 2) * pl.col("yaw").sin()
                        + (+pl.col("width") / 2) * pl.col("yaw").cos()
                    ).alias("y4"),
                )
                .to_numpy()
            )
            polys = shapely.polygons(np.stack([ar[:, :4], ar[:, 4:]], axis=2))
            df = df.with_columns(pl.Series(name="polygon", values=polys))
        return df

    def _add_polygons_to_df(self):
        self._df = self._add_polygons(self._df)

    @staticmethod
    def get_moving_object_ground_truth(
        nanos: int, df: pl.DataFrame, host_vehicle=None, validate=False
    ) -> betterosi.GroundTruth:
        if validate:
            recording_moving_object_schema.validate(df.to_pandas(), lazy=True)

        def get_object(row):
            return betterosi.MovingObject(
                id=betterosi.Identifier(value=row["idx"]),
                type=betterosi.MovingObjectType(row["type"]),
                base=betterosi.BaseMoving(
                    dimension=betterosi.Dimension3D(length=row["length"], width=row["width"], height=row["width"]),
                    position=betterosi.Vector3D(x=row["x"], y=row["y"], z=row["z"]),
                    orientation=betterosi.Orientation3D(roll=row["roll"], pitch=row["pitch"], yaw=row["yaw"]),
                    velocity=betterosi.Vector3D(x=row["vel_x"], y=row["vel_y"], z=row["vel_z"]),
                    acceleration=betterosi.Vector3D(x=row["acc_x"], y=row["acc_y"], z=row["acc_z"]),
                ),
                vehicle_classification=betterosi.MovingObjectVehicleClassification(
                    type=row["subtype"], role=row["role"]
                ),
            )

        mvs = [get_object(r) for r in df.iter_rows(named=True)]
        gt = betterosi.GroundTruth(
            version=betterosi.InterfaceVersion(version_major=3, version_minor=7, version_patch=9),
            timestamp=betterosi.Timestamp(seconds=int(nanos // 1_000_000_000), nanos=int(nanos % 1_000_000_000)),
            host_vehicle_id=betterosi.Identifier(value=0)
            if host_vehicle is None
            else betterosi.Identifier(value=host_vehicle),
            moving_object=mvs,
        )
        return gt

    def __init__(self, df, map=None, projections=None, host_vehicle=None, validate=False, compute_polygons=False):
        if validate:
            recording_moving_object_schema.validate(df.to_pandas() if isinstance(df, pl.DataFrame) else df, lazy=True)
        if not isinstance(df, pl.DataFrame):
            df = pl.DataFrame(df)
        super().__init__()
        self.nanos2frame = {n: i for i, n in enumerate(df["total_nanos"].unique())}
        mapping = pl.DataFrame({"total_nanos": list(self.nanos2frame.keys()), "frame": list(self.nanos2frame.values())})
        if "frame" in df.columns:
            df = df.drop("frame")
        df = df.join(mapping, on="total_nanos", how="left")
        if "polygon" not in df.columns and compute_polygons:
            df = self._add_polygons(df)
        if "vel" not in df.columns:
            df = df.with_columns(
                pl.struct(["vel_x", "vel_y"])
                .map_batches(partial(np.linalg.norm, axis=1), is_elementwise=True)
                .alias("vel")
            )
        if "acc" not in df.columns:
            df = df.with_columns(
                pl.struct(["acc_x", "acc_y"])
                .map_batches(partial(np.linalg.norm, axis=1), is_elementwise=True)
                .alias("acc")
            )
        self.projections = projections if projections is not None else []
        self._df = df
        self.map = map
        self._moving_objects = (
            None  # = {int(idx): self._MovingObjectClass(self, idx) for idx in self._df["idx"].unique()}
        )
        self.host_vehicle = host_vehicle

    @property
    def moving_objects(self):
        if self._moving_objects is None:
            self._moving_objects = {int(idx): self._MovingObjectClass(self, idx) for idx in self._df["idx"].unique()}
        return self._moving_objects

    def to_osi_gts(self) -> list[betterosi.GroundTruth]:
        first_iteration = True
        for [nanos], group_df in self._df.sort(["total_nanos"]).group_by("total_nanos", maintain_order=True):
            gt = self.get_moving_object_ground_truth(nanos, group_df, host_vehicle=self.host_vehicle, validate=False)
            if first_iteration:
                first_iteration = False
                if self.map is not None and isinstance(self.map, MapOsi | MapOsiCenterline):
                    gt.lane_boundary = [b._osi for b in self.map.lane_boundaries.values()]
                    gt.lane = [l._osi for l in self.map.lanes.values()]
            yield gt

    @classmethod
    def from_osi_gts(cls, gts: list[betterosi.GroundTruth], **kwargs):
        projs = []

        def get_gts():
            for i, gt in enumerate(gts):
                total_nanos = gt.timestamp.seconds * 1_000_000_000 + gt.timestamp.nanos
                if gt.proj_frame_offset is not None and gt.proj_frame_offset.position is None:
                    raise ValueError(
                        f"Offset of {i}th ground truth message (total_nanos={total_nanos}) is set without position."
                    )
                projs.append(
                    dict(
                        proj_string=gt.proj_string,
                        offset=ProjectionOffset(
                            x=gt.proj_frame_offset.position.x,
                            y=gt.proj_frame_offset.position.y,
                            z=gt.proj_frame_offset.position.z,
                            yaw=gt.proj_frame_offset.yaw,
                        )
                        if gt.proj_frame_offset is not None
                        else None,
                    )
                )
                for mv in gt.moving_object:
                    yield dict(
                        total_nanos=total_nanos,
                        idx=mv.id.value,
                        x=mv.base.position.x,
                        y=mv.base.position.y,
                        z=mv.base.position.z,
                        vel_x=mv.base.velocity.x,
                        vel_y=mv.base.velocity.y,
                        vel_z=mv.base.velocity.z,
                        acc_x=mv.base.acceleration.x,
                        acc_y=mv.base.acceleration.y,
                        acc_z=mv.base.acceleration.z,
                        length=mv.base.dimension.length,
                        width=mv.base.dimension.width,
                        height=mv.base.dimension.height,
                        roll=mv.base.orientation.roll,
                        pitch=mv.base.orientation.pitch,
                        yaw=mv.base.orientation.yaw,
                        type=mv.type,
                        role=mv.vehicle_classification.role
                        if mv.type == betterosi.MovingObjectType.TYPE_VEHICLE
                        else -1,
                        subtype=mv.vehicle_classification.type
                        if mv.type == betterosi.MovingObjectType.TYPE_VEHICLE
                        else -1,
                    )

        df_mv = pl.DataFrame(get_gts()).sort(["total_nanos", "idx"])
        return cls(df_mv, projections=projs, **kwargs)

    @classmethod
    def from_file(
        cls,
        filepath,
        xodr_path: str | None = None,
        validate: bool = False,
        parse_map: bool = False,
        compute_polygons: bool = False,
    ):
        if Path(filepath).suffix == ".parquet":
            return cls.from_parquet(filepath, parse_map=parse_map, validate=validate, compute_polygons=compute_polygons)

        gts = betterosi.read(filepath, return_ground_truth=True, mcap_return_betterosi=False)
        gts, tmp_gts = itertools.tee(gts, 2)
        first_gt = next(tmp_gts)
        r = cls.from_osi_gts(gts, validate=validate, compute_polygons=compute_polygons)
        if xodr_path is not None:
            r.map = MapOdr.from_file(xodr_path, parse=parse_map)
        elif Path(filepath).suffix == ".mcap":
            try:
                r.map = MapOdr.from_file(filepath, parse=parse_map)
            except StopIteration:
                pass
        if r.map is None:
            try:
                r.map = MapOsi.create(first_gt)
                warn("No map provided in mcap. OSI GroundTruth map is used!")
            except RuntimeError:
                try:
                    r.map = MapOsiCenterline.create(first_gt)
                    warn("No map provided in mcap. OSI GroundTruth (Centerline) map is used!")
                except RuntimeError:
                    pass
        if r.map is None:
            warn("No xodr map provided in MCAP nor OSI map in GroundTruth!")
        return r

    def to_mcap(self, filepath):
        if Path(filepath).suffix != ".mcap":
            raise ValueError()
        with betterosi.Writer(filepath) as w:
            for gt in self.to_osi_gts():
                w.add(gt)
            if isinstance(self.map, MapOdr):
                w.add(self.map.to_osi(), topic="ground_truth_map", log_time=0)

    def to_hdf(self, filename, key="moving_object"):
        #!pip install tables
        to_drop = [] if "polygon" not in self._df.columns else ["polygon"]
        to_drop += ["frame"]
        self._df.drop(columns=to_drop).to_pandas().to_hdf(filename, key=key)

    @classmethod
    def from_hdf(cls, filename, key="moving_object"):
        df = pl.DataFrame(pd.read_hdf(filename, key=key))
        return cls(df, map=None, host_vehicle=None)

    def interpolate(self, new_nanos: list[int] | None = None, hz: float | None = None):
        df = self._df
        if new_nanos is None and hz is None:
            new_nanos = np.linspace(
                df["total_nanos"].min(), df["total_nanos"].max(), df["frame"].max() - df["frame"].min(), dtype=int
            )
        elif hz is not None:
            step = 1_000_000_000 / hz
            new_nanos = np.arange(start=df["total_nanos"].min(), stop=df["total_nanos"].max() + 1, step=step, dtype=int)
        else:
            new_nanos = np.array(new_nanos)
        new_dfs = []
        for [idx], track_df in df.group_by("idx"):
            track_data = {}
            track_new_nanos = new_nanos[
                track_df["frame"].min() - df["frame"].min() : track_df["frame"].max() - df["frame"].min() + 1
            ]
            for c in ["x", "y", "z", "vel_x", "vel_y", "vel_z", "acc_x", "acc_y", "acc_z", "length", "width", "height"]:
                track_data[c] = np.interp(track_new_nanos, track_df["total_nanos"], track_df[c])
            for c in ["type", "subtype", "role"]:
                track_data[c] = nearest_interp(
                    track_new_nanos, track_df["total_nanos"].to_numpy(), track_df[c].to_numpy()
                )
            for c in ["roll", "pitch", "yaw"]:
                track_data[c] = np.mod(
                    np.interp(track_new_nanos, track_df["total_nanos"], np.unwrap(track_df[c], period=np.pi)), np.pi
                )
            new_track_df = pl.DataFrame(track_data)
            new_track_df = new_track_df.with_columns(
                pl.Series(name="idx", values=np.ones_like(track_new_nanos) * idx),
                pl.Series(name="total_nanos", values=track_new_nanos),
            )
            new_dfs.append(new_track_df)
        new_df = pl.concat(new_dfs)
        return self.__init__(new_df, self.map, self.host_vehicle)

    def plot(self, ax=None, legend=False) -> plt.Axes:
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        if self.map:
            self.map.plot(ax)
        self.plot_mvs(ax=ax)
        if legend:
            ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
        return ax

    def plot_mvs(self, ax=None, legend=False):
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        for [idx], mv in self._df["idx", "x", "y"].group_by("idx"):
            ax.plot(mv["x"], mv["y"], c="red", alpha=0.5, label=str(idx))
        if legend:
            ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
        return ax

    def plot_frame(self, frame: int, ax=None):
        ax = self.plot(ax=ax)
        self.plot_mv_frame(ax, frame=frame)
        return ax

    def plot_mv_frame(self, ax: plt.Axes, frame: int):
        if "polygon" not in self._df.columns:
            self._add_polygons_to_df()
        polys = self._df.filter(pl.col("frame") == frame)["polygon"]
        for p in polys:
            ax.add_patch(PltPolygon(p.exterior.coords, fc="red"))

    @classmethod
    def from_parquet(cls, filename, parse_map: bool = False, **kwargs):
        t = pq.read_table(filename)
        df = pl.DataFrame(t)
        if t.schema.metadata is not None and b"xodr" in t.schema.metadata:
            m = MapOdr.create(
                odr_xml=t.schema.metadata[b"xodr"].decode(),
                name=t.schema.metadata[b"xodr_name"].decode(),
                parse=parse_map,
            )
        else:
            m = None
        return cls(df, map=m, **kwargs)

    def to_parquet(self, filename):
        if len(self.projections) > 0:
            proj_dict = {
                b"proj_string": self.projections[0]["proj_string"],
            }
            if self.projections[0]["offset"] is not None:
                proj_dict[b"offset_x"] = str(self.projections[0]["offset"].x).encode()
                proj_dict[b"offset_y"] = str(self.projections[0]["offset"].y).encode()
                proj_dict[b"offset_z"] = str(self.projections[0]["offset"].z).encode()
                proj_dict[b"offset_yaw"] = str(self.projections[0]["offset"].yaw).encode()

        else:
            proj_dict = {}
        to_drop = ["frame"]
        if "polygon" in self._df.columns:
            to_drop.append("polygon")
        t = pyarrow.table(self._df.drop(*to_drop))
        if hasattr(self.map, "odr_xml"):
            t = t.cast(
                t.schema.with_metadata(
                    proj_dict | {b"xodr": self.map.odr_xml.encode(), b"xodr_name": self.map.name.encode()}
                )
            )
        else:
            t = t.cast(t.schema.with_metadata(proj_dict | {}))
        pq.write_table(t, filename)

    def plot_altair(self, start_frame=0, end_frame=-1, plot_map=True, metric_column=None, idx=None):
        if "polygon" not in self._df.columns:
            self._df = self._add_polygons(self._df)
        if "geometry" not in self._df.columns:
            self._df = self._df.with_columns(geometry=st.from_shapely("polygon"))
        if not hasattr(self, "_map_df") and plot_map:
            self._map_df = pl.DataFrame(
                [
                    pl.Series(name="polygon", values=[l.polygon for l in self.map.lanes.values()]),
                    pl.Series(name="idx", values=[i for i, _ in enumerate(self.map.lanes.keys())]),
                    pl.Series(name="type", values=[o.type.name for o in self.map.lanes.values()]),
                ]
            )
            self._map_df = self._map_df.with_columns(geometry=st.from_shapely("polygon"))

        if end_frame != -1:
            df = self._df.filter(pl.col("frame") < end_frame, pl.col("frame") >= start_frame)
        else:
            df = self._df.filter(pl.col("frame") >= start_frame)

        [frame_min], [frame_max] = df.select(
            pl.col("frame").min().alias("min"),
            pl.col("frame").max().alias("max"),
        )[0]
        slider = alt.binding_range(min=frame_min, max=frame_max, step=1, name="frame")
        op_var = alt.param(value=0, bind=slider)

        [xmin], [xmax], [ymin], [ymax] = df.select(
            pl.col("x").min().alias("xmin"),
            pl.col("x").max().alias("xmax"),
            pl.col("y").min().alias("ymin"),
            pl.col("y").max().alias("ymax"),
        )[0]
        pov = {
            "type": "Feature",
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[xmax, ymax], [xmax, ymin], [xmin, ymin], [xmin, ymax], [xmax, ymax]]],
            },
            "properties": {},
        }
        map = (
            self._map_df["geometry", "idx", "type"]
            .st.plot(color="green", fillOpacity=0.4)
            .encode(tooltip=["properties.idx:N", "properties.type:O"])
        )
        mvs = (
            df["geometry", "idx", "frame", "type"]
            .st.plot()
            .encode(
                tooltip=["properties.idx:N", "properties.frame:N", "properties.type:O"],
                color=alt.value("blue")
                if idx is None
                else alt.when(alt.FieldEqualPredicate(equal=idx, field="properties.idx"))
                .then(alt.value("red"))
                .otherwise(alt.value("blue")),
            )
            .transform_filter(alt.FieldEqualPredicate(field="properties.frame", equal=op_var))
        )

        map_view = (
            (map + mvs)
            .project("identity", reflectY=True, fit=pov)
            .properties(height=int(ymax - ymin) * 3, width=int(xmax - xmin) * 3, title="Map")
        )
        view = map_view
        if metric_column is not None and idx is not None:
            metric = (
                df["idx", metric_column, "frame"]
                .filter(idx=idx)
                .plot.line(x="frame", y=metric_column, color=alt.value("red"))
                .properties(title=f"{metric_column} of object {idx}")
            )
            vertline = (
                alt.Chart()
                .mark_rule()
                .encode(x=alt.datum(op_var, type="quantitative", scale=alt.Scale(domain=[frame_min, frame_max])))
            )
            view = view | (metric + vertline)
        return view.add_params(op_var)



import cProfile
import json
import numpy as np
import shapely
from pathlib import Path
from pstats import Stats


import omega_prime

p = Path(__file__).parent.parent / "example_files/"
with open(p / "mapping.json") as f:
    mapping = json.load(f)


def test_validation():
    with cProfile.Profile() as pr:
        rec = omega_prime.Recording.from_file(p / mapping[3][0], p / mapping[3][1])
        rec.to_mcap("validated.mcap")
        rec = omega_prime.Recording.from_file("validated.mcap", validate=True)
        stats = Stats(pr)
    stats.dump_stats("test_validate.prof")


def test_esmini_examples():
    with cProfile.Profile() as pr:
        for p_osi, p_odr in mapping:
            rec = omega_prime.Recording.from_file(p / p_osi, p / p_odr, validate=False)
            rec.to_mcap(f"{Path(p_osi).stem}.mcap")
            rec = omega_prime.Recording.from_file(f"{Path(p_osi).stem}.mcap", validate=False)
        stats = Stats(pr)
    stats.dump_stats("test.prof")


def test_interpolate():
    rec = omega_prime.Recording.from_file(p / mapping[3][0], p / mapping[3][1], validate=False)
    rec.interpolate(hz=10)


def test_centerline():
    with cProfile.Profile() as pr:
        # Load the recording
        rec = omega_prime.Recording.from_file(p / mapping[3][0], p / mapping[3][1], validate=False)

        # Create a locator and remove all polygons from the lanes
        locator = omega_prime.Locator.from_mapodr(rec.map)
        for lane in locator.all_lanes:
            lane.polygon = None  # Remove polygons

        # Create a new locator using only centerlines
        locator_cl = omega_prime.Locator(locator.all_lanes, locator.internal2external_laneid)

        # Prepare input coordinates for xys2sts
        xys = np.stack([rec.moving_objects[0].x, rec.moving_objects[0].y]).T

        # Run xys2sts and ensure it executes without errors
        sts_cl = locator_cl.xys2sts(xys)

        # Check that str_tree.geometries only contains LineString objects
        assert all(isinstance(geom, shapely.LineString) for geom in locator_cl.str_tree.geometries), (
            "str_tree contains non-LineString geometries"
        )

        # check that sts_cl is not empty or has expected properties
        assert len(sts_cl["s"]) > 0, "sts_cl output is empty"
        assert len(sts_cl["t"]) > 0, "sts_cl output is empty"

    # Save profiling results to a file
    stats = Stats(pr)
    stats.dump_stats("test_centerline.prof")


def test_locator():
    with cProfile.Profile() as pr:
        # Load the recording
        rec = omega_prime.Recording.from_file(p / mapping[3][0], p / mapping[3][1], validate=False)

        # Create a locator and remove all polygons from the lanes
        locator = omega_prime.Locator.from_mapodr(rec.map)

        # Prepare input coordinates for xys2sts
        xys = np.stack([rec.moving_objects[0].x, rec.moving_objects[0].y]).T

        # Run xys2sts and ensure it executes without errors
        sts_cl = locator.xys2sts(xys)

        # check that sts_cl is not empty or has expected properties
        assert len(sts_cl["s"]) > 0, "sts_cl output is empty"
        assert len(sts_cl["t"]) > 0, "sts_cl output is empty"

    # Save profiling results to a file
    stats = Stats(pr)
    stats.dump_stats("test_centerline.prof")


def test_parquet():
    rec = omega_prime.Recording.from_file(p / mapping[3][0], p / mapping[3][1])
    rec.to_parquet("test.parquet")
    rec = omega_prime.Recording.from_file("test.parquet")
    assert rec.map is not None


if __name__ == "__main__":
    test_locator()
    # pass



__pdoc__ = {}
__pdoc__["converters"] = False
""" .. include:: ./../README.md """
from . import converters
from .map_odr import MapOdr
from .locator import LaneRelation, Locator
from .map import Lane, LaneBoundary, Map, MapOsi
from .recording import MovingObject, Recording
from .converters import convert_lxd

__all__ = [
    "Recording",
    "MovingObject",
    "MapOsi",
    "Map",
    "Lane",
    "LaneBoundary",
    "MapOdr",
    "Locator",
    "LaneRelation",
    "converters",
    "convert_lxd",
]



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any

import betterosi
import networkx as nx
import numpy as np
import shapely
import xarray as xr
from matplotlib.patches import Polygon as PltPolygon
from strenum import StrEnum


class ShapelyTrajectoryTools:
    epsi = 1e-10
    l_append = 200
    simplify_tolerance = 1e-5

    @staticmethod
    def plot_polygon(ax, p, **kwargs):
        if not p.is_empty:
            if isinstance(p, shapely.geometry.MultiPolygon):
                ps = p.geoms
            else:
                ps = [p]
            for p in ps:
                try:
                    ax.add_patch(PltPolygon(p.exterior.coords, **kwargs))
                except AttributeError as e:
                    raise e
                    # p is not a polygon
                    pass

    @classmethod
    def extend_linestring(cls, l: shapely.LineString, simplify: bool = True, l_append=None):
        l_append = l_append if l_append is not None else cls.l_append
        # TODO move s coordinate epsi to NOT normalized
        if isinstance(l, shapely.Point):
            cl = np.asarray(l.coords)[0]
            return shapely.LineString(np.stack([cl - np.array([0, 1]), cl, cl + np.array([0, 1])]))

        startvec = np.asarray(l.interpolate(0, normalized=True).coords) - np.asarray(
            l.interpolate(0 + cls.epsi, normalized=True).coords
        )
        endvec = np.asarray(l.interpolate(1, normalized=True).coords) - np.asarray(
            l.interpolate(1 - cls.epsi, normalized=True).coords
        )
        if np.all(startvec == 0):
            startvec = np.asarray(l.interpolate(0, normalized=True).coords) - np.asarray(
                l.interpolate(1, normalized=True).coords
            )
        if np.all(endvec == 0):
            endvec = np.asarray(l.interpolate(1, normalized=True).coords) - np.asarray(
                l.interpolate(0, normalized=True).coords
            )

        startvec = startvec / np.linalg.norm(startvec) * l_append
        endvec = endvec / np.linalg.norm(endvec) * l_append
        cl = np.asarray(l.coords)
        cl = shapely.LineString(np.concatenate([startvec + cl[0, :], cl, endvec + cl[-1, :]]))
        return cl.simplify(tolerance=cls.simplify_tolerance) if simplify else cl

    @classmethod
    def get_linestring_coordinate_s(cls, l: shapely.LineString):
        return shapely.line_locate_point(l, shapely.points(l.coords))

    @classmethod
    def st2xy(cls, l: shapely.LineString, s, t, return_heading_of_ref_at_st=False):
        ref_points_after = l.interpolate(np.clip(s + cls.epsi, 0, l.length))
        ref_points = l.interpolate(np.clip(s, 0, l.length))
        ref_points_before = l.interpolate(np.clip(s - cls.epsi, 0, l.length))
        lane_tangent_vec = np.array([np.array(o.coords)[0] for o in ref_points_after]) - np.array(
            [np.array(o.coords)[0] for o in ref_points_before]
        )
        lane_tangent_norm_angle = np.arctan2(lane_tangent_vec[:, 1], lane_tangent_vec[:, 0]) - np.pi / 2
        if return_heading_of_ref_at_st:
            return lane_tangent_norm_angle + np.pi / 2
        else:
            lane_tangent_complex_vec = np.exp(1j * lane_tangent_norm_angle) * t
            ref_points = np.array([np.array(o.coords)[0] for o in ref_points])
            x = np.real(lane_tangent_complex_vec) + ref_points[:, 0]
            y = np.imag(lane_tangent_complex_vec) + ref_points[:, 1]
            return x, y

    @classmethod
    def needs_angle_adjustment(cls, lane_point_distance, lon_distances):
        lpd = lane_point_distance[:, np.newaxis]
        ub = lpd + cls.epsi
        lb = lpd - cls.epsi
        in_corner_area = np.logical_and(ub <= lon_distances, lb <= lon_distances)
        return np.any(in_corner_area, axis=0)

    @classmethod
    def xy2st(cls, l: shapely.LineString, x_or_xy, y=None, line_point_distances=None):
        if y is None:
            xy_points = x_or_xy
        else:
            xy_points = shapely.points(np.stack([x_or_xy, y], axis=1))
        lon_distances = l.project(xy_points)
        is_driver_side_of_centerline = shapely.is_ccw(
            shapely.linearrings(
                np.array(
                    [
                        [np.asarray(o.coords)[0, :2] for o in l.interpolate(np.clip(lon_distances - 0.1, 0, np.inf))],
                        [np.asarray(o.coords)[0, :2] for o in l.interpolate(np.clip(lon_distances + 0.1, 0, l.length))],
                        [np.asarray(o.coords)[0, :2] for o in xy_points],
                    ]
                ).transpose(1, 0, 2)
            )
        )
        lat_distances = l.distance(xy_points) * (-is_driver_side_of_centerline.astype(int) * 2 + 1)

        delta_s = np.zeros_like(lon_distances)
        if line_point_distances is None:
            lane_point_distances = cls.get_linestring_coordinate_s(l)
        else:
            lane_point_distances = line_point_distances
        delta_s_idxs = cls.needs_angle_adjustment(lane_point_distances, lon_distances)
        if np.sum(delta_s_idxs) > 0:
            points = np.stack([p.centroid.coords for p in xy_points[delta_s_idxs]])[:, 0, :2]
            lon_dist_to_fix = lon_distances[delta_s_idxs]
            before_nearest_points = np.stack(
                [np.asarray(o.coords)[0, :2] for o in l.interpolate(np.clip(lon_dist_to_fix - cls.epsi, 0, l.length))]
            )
            nearest_points = np.stack(
                [np.asarray(o.coords)[0, :2] for o in l.interpolate(np.clip(lon_dist_to_fix, 0, l.length))]
            )
            after_nearest_points = np.stack(
                [np.asarray(o.coords)[0, :2] for o in l.interpolate(np.clip(lon_dist_to_fix + cls.epsi, 0, l.length))]
            )

            before_vec = nearest_points - before_nearest_points
            before_angle = np.arctan2(before_vec[:, 1], before_vec[:, 0])
            after_vec = after_nearest_points - nearest_points
            after_angle = np.arctan2(after_vec[:, 1], after_vec[:, 0])
            point_vec = points - nearest_points
            point_angle = np.arctan2(point_vec[:, 1], point_vec[:, 0])
            normal_angle = point_angle - np.pi / 2
            before_in_angle = np.mod((before_angle - normal_angle), np.pi)
            after_in_angle = np.mod(normal_angle + np.pi - after_angle, np.pi)
            aia_sin = np.sin(after_in_angle)
            iz = np.isclose(aia_sin, 0)
            gamma = np.ones_like(before_in_angle) * np.inf
            gamma[~iz] = np.sin(before_in_angle[~iz]) / aia_sin[~iz]
            after_length = 2 * cls.epsi * (1 - (1 / (gamma + 1)))
            delta_s[delta_s_idxs] = cls.epsi - after_length
        return np.stack([lon_distances - delta_s, lat_distances]).T


def get_lane_centerline(right_border: shapely.LineString, left_border: shapely.LineString) -> shapely.LineString:
    """middle line between (interpolated) boundaries, oriented in direction of lane"""
    ses = np.unique(
        np.concatenate(
            [
                shapely.line_locate_point(left_border, shapely.points(right_border.coords), normalized=True),
                shapely.line_locate_point(right_border, shapely.points(left_border.coords), normalized=True),
            ]
        )
    )

    points = np.zeros((len(ses), 2))
    for i, (rbp, lbp) in enumerate(
        zip(right_border.interpolate(ses, normalized=True), left_border.interpolate(ses, normalized=True))
    ):
        points[i, :] = shapely.MultiPoint([rbp, lbp]).minimum_rotated_rectangle.centroid.coords

    # TODO some smoothing operation could be helpful
    cl = shapely.LineString(points)

    if cl.is_empty or not cl.is_valid:
        raise RuntimeError("Could not compute centerline for lane!")
    return cl


class LaneRelation(StrEnum):
    predecessor = "predecessor"
    successor = "successor"
    neighbour_right = "right neighbour"
    neighbour_left = "left neighbour"


@dataclass
class MockRoad:
    odr_idx: int


@dataclass
class WrappedLane:
    idx: Any
    road: Any
    centerline: shapely.LineString
    polygon: shapely.Polygon
    border_left: shapely.LineString
    border_right: shapely.LineString
    predecessor_idxs: list[Any]
    successor_idxs: list[Any]
    start_points: list[shapely.Point]
    end_points: list[shapely.Point]
    on_intersection: bool = False
    type: Any = None
    original_lane: Any = None
    db: Any = None

    @classmethod
    def from_odr_lane(cls, map, l):
        start_points = np.stack([l.left_boundary.polyline[0, :2], l.right_boundary.polyline[0, :2]])
        end_points = np.stack([l.left_boundary.polyline[-1, :2], l.right_boundary.polyline[-1, :2]])
        l.odr_idx = l.idx[1]
        l.road = MockRoad(odr_idx=l.idx[0])
        return cls(
            idx=l.idx,
            road=l.road,
            centerline=shapely.LineString(l.centerline[:, :2]),
            polygon=l.polygon,
            border_left=shapely.LineString(l.left_boundary.polyline[:, :2]),
            border_right=shapely.LineString(l.right_boundary.polyline[:, :2]),
            predecessor_idxs=l.predecessor_ids,
            successor_idxs=l.successor_ids,
            start_points=start_points,
            end_points=end_points,
            on_intersection=l.type == betterosi.LaneClassificationType.TYPE_INTERSECTION,
            type=l.type,
            original_lane=l,
        )


@dataclass
class Locator:
    all_lanes: Any  # array of all lanes
    external2internal_laneid: dict[Any, int] = field(init=False)
    internal2external_laneid: list[Any] = field(init=False)
    lane_point_distances: list = field(init=False)
    str_tree: shapely.STRtree = field(init=False)
    extended_centerlines: list[shapely.LineString] = field(init=False)

    g: nx.DiGraph = field(init=False)  # Lane Relation Graph
    xodr2external_laneid: None | dict = None

    @classmethod
    def from_mapodr(cls, map):
        all_lanes = [WrappedLane.from_odr_lane(map, l) for l in map.lanes.values()]
        return cls(all_lanes=all_lanes, xodr2external_laneid={l.idx: l.idx for l in all_lanes})

    def __post_init__(self):
        self.external2internal_laneid = {l.idx: i for i, l in enumerate(self.all_lanes)}
        self.extended_centerlines = [ShapelyTrajectoryTools.extend_linestring(l.centerline) for l in self.all_lanes]
        if hasattr(self.all_lanes[0], "polygon") and self.all_lanes[0].polygon is not None:
            self.str_tree = shapely.STRtree([l.polygon for l in self.all_lanes])
        else:
            self.str_tree = shapely.STRtree([l.centerline for l in self.all_lanes])
        self.internal2external_laneid = [l.idx for l in self.all_lanes]
        self.lane_point_distances = [
            np.unique(shapely.line_locate_point(cl, shapely.points(cl.coords))) for cl in self.extended_centerlines
        ]
        self.g = self._get_routing_graph()

    def fullid2xodrid(self, roadlane_id):
        """
        conversion of roadlane id to xodr ids
        """
        if isinstance(roadlane_id[0], list):
            roadlane_id = tuple([tuple(roadlane_id[0]), tuple(roadlane_id[1])])
        else:
            roadlane_id = tuple(roadlane_id)
        l = self.all_lanes[self.external2internal_laneid[roadlane_id]]
        return l.road.odr_idx, l.odr_idx

    def get_route(self, start_id, end_id):
        return nx.shortest_path(self.g, start_id, end_id)

    def sts2xys(self, sts):
        xys = np.zeros((len(sts.s), 2), dtype=float) * np.nan
        l_ids = np.array([self.external2internal_laneid[i] for i in sts.roadlane_id.values])
        for l_id in set(l_ids):
            point_idxs = np.argwhere(l_ids == l_id)[:, 0]
            rel_sts = sts.isel(dict(time=point_idxs))
            l = self.extended_centerlines[l_id]
            xys[point_idxs, 0], xys[point_idxs, 1] = ShapelyTrajectoryTools.st2xy(
                l, rel_sts.s.values + ShapelyTrajectoryTools.l_append, rel_sts.t.values
            )
        return xys

    def xys2sts(self, xys, poly=None):
        if isinstance(xys, np.ndarray) and xys.ndim == 2:
            assert xys.shape[1] == 2
            xys = shapely.points(xys)
        lat_distances, lon_distances = self._xys2sts(xys, poly)
        single_lane_association = self.get_single_lane_association(lat_distances)
        sla = np.zeros(len(single_lane_association), dtype=tuple)
        for i, v in enumerate(single_lane_association):
            sla[i] = v
        sts = xr.Dataset(
            {
                "s": ("time", [lon_distances[lidx][i] for i, lidx in enumerate(single_lane_association)]),
                "t": ("time", [lat_distances[lidx][i] for i, lidx in enumerate(single_lane_association)]),
                "roadlane_id": ("time", sla),
            }
        )
        return sts

    def xys2lane_sts(self, lane_id, xys, internal_id=False):
        # xys should be an array of shapely objects or an array of points with dim (n_points, 2)
        # return (n_points, 2) where ret[:,0] is s and ret[:,1] is t
        if isinstance(xys, np.ndarray) and xys.ndim == 2:
            assert xys.shape[1] == 2
            xys = shapely.points(xys)
        lid = self.external2internal_laneid[lane_id] if not internal_id else lane_id
        lane_point_distances = self.lane_point_distances[lid]
        sts = ShapelyTrajectoryTools.xy2st(
            self.extended_centerlines[lid], x_or_xy=xys, line_point_distances=lane_point_distances
        )
        sts[:, 0] -= ShapelyTrajectoryTools.l_append
        return sts

    def _xys2sts(self, xys, polys=None):
        # xys should be an array of shapely objects or an array of points with dim (n_points, 2)
        if isinstance(xys, np.ndarray) and xys.ndim == 2:
            assert xys.shape[1] == 2
            xys = shapely.points(xys)
        else:
            xys = np.array(xys)
        if polys is None:
            polys = xys
        lon_distances = defaultdict(lambda: np.nan * np.ones((len(xys),)))
        lat_distances = defaultdict(lambda: np.nan * np.ones((len(xys),)))
        point_idxs, intersection_lane_ids = self.str_tree.query(polys, predicate="intersects")
        for l_id in set(intersection_lane_ids):
            lps = point_idxs[intersection_lane_ids == l_id]
            (
                lon_distances[self.internal2external_laneid[l_id]][lps],
                lat_distances[self.internal2external_laneid[l_id]][lps],
            ) = self.xys2lane_sts(l_id, xys[lps], internal_id=True).T
        try:
            no_associations = np.where(np.all(np.isnan(np.stack(list(lon_distances.values()))), axis=0))[0]
        except ValueError:
            # no arrays to stack
            no_associations = np.array(list(range(len(xys))))
        if hasattr(self.all_lanes[0], "polygon") and self.all_lanes[0].polygon is not None:
            no_asscociation_idxs, intersection_lane_ids = self.str_tree.query_nearest(polys[no_associations])
        else:
            # Create an empty numpy array for no_asscociation_idxs
            no_asscociation_idxs = np.array([])
            intersection_lane_ids = np.array([])
            if len(no_associations) > 0:
                for idx, poly in enumerate(polys[no_associations]):
                    # Returns the indxes of all centerlines that are in range
                    nearby_idx = self.query_centerlines(poly, range_percentage=0.1)
                    # Connect the no_assosciation_idxs with the intersection_lane_ids
                    no_asscociation_idxs = np.append(no_asscociation_idxs, [no_associations[idx]] * len(nearby_idx))
                    intersection_lane_ids = np.append(intersection_lane_ids, nearby_idx)

        # Need a convertion from float values to int values. This is because the shapely STRtree query_nearest returns float values
        no_asscociation_idxs = no_asscociation_idxs.astype(int)
        intersection_lane_ids = intersection_lane_ids.astype(int)
        for l_id in set(intersection_lane_ids):
            lps = no_associations[no_asscociation_idxs[intersection_lane_ids == l_id]]
            (
                lon_distances[self.internal2external_laneid[l_id]][lps],
                lat_distances[self.internal2external_laneid[l_id]][lps],
            ) = self.xys2lane_sts(l_id, xys[lps], internal_id=True).T

        no_asscociation_new = np.where(np.all(np.isnan(np.stack(list(lon_distances.values()))), axis=0))[0]

        assert len(no_asscociation_new) == 0
        return lat_distances, lon_distances

    def query_centerlines(self, point, range_percentage=0.1):
        """
        Query the nearest centerline and all centerlines within a range percentage.

        :param point: A shapely Point object representing the query location.
        :param range_percentage: The range as a percentage of the total length of the nearest centerline. Default is 0.1 (10%).
        :return: A NDArray with all the Lane Idx in the Range.
        """
        # Query the nearest centerline
        nearest_idx = self.str_tree.query_nearest(point)
        nearest_centerline = self.extended_centerlines[nearest_idx[0]]

        # Calculate the range based on the nearest centerline's length
        range_distance = nearest_centerline.distance(point) * (1 + range_percentage)

        # Create a buffer around the point
        buffer = point.buffer(range_distance)

        # Query all centerlines within the buffer
        nearby_idxs = self.str_tree.query(buffer, predicate="intersects")

        # If there was no intersection, return the nearest centerline
        if nearby_idxs.size == 0:
            return nearest_idx

        return nearby_idxs

    def _get_routing_graph(self):
        all_lanes = self.all_lanes
        str_tree = self.str_tree
        external2internal_laneid = self.external2internal_laneid
        g = nx.DiGraph()
        for lid, lane in enumerate(all_lanes):
            for external_pid in lane.predecessor_idxs:
                try:
                    g.add_edge(lid, external2internal_laneid[external_pid], label=LaneRelation.predecessor)
                except KeyError:
                    pass
            for external_sid in lane.successor_idxs:
                try:
                    g.add_edge(lid, external2internal_laneid[external_sid], label=LaneRelation.successor)
                except KeyError:
                    pass
            right_neigbours = [i for i in str_tree.query(lane.border_right, predicate="covered_by") if i != lid]
            left_neigbours = [i for i in str_tree.query(lane.border_left, predicate="covered_by") if i != lid]
            for rn in right_neigbours:
                g.add_edge(lid, rn)
            for ln in left_neigbours:
                g.add_edge(lid, ln)
        return g

    def get_single_lane_association(
        self, traveler_lane_intersections: dict[Any, Any], overlaps: None | dict[Any, float] = None
    ):
        """
        filter traveling path of traveler, so that traveler is not assigned to lanes that are only reachable through a merging or crossing relation
        return format: road, lane
        """
        import networkx as nx

        g = nx.Graph()
        nodes = defaultdict(list)
        for external_lid, v in traveler_lane_intersections.items():
            for timeidx in np.where(~np.isnan(v))[0]:
                nodes[timeidx].append(self.external2internal_laneid[external_lid])
        nodes_per_time = [nodes[i] for i in range(len(nodes))]
        g.add_node("start", pos=(-1, -1))
        g.add_node("end", pos=(len(nodes_per_time), -1))
        for i, nodes_of_time in enumerate(nodes_per_time[:-1]):
            for n in nodes_of_time:
                for next_n in nodes_per_time[i + 1]:
                    g.add_node((n, i), pos=(i, n))
                    g.add_node((next_n, i + 1), pos=(i + 1, next_n))
                    try:
                        if n == next_n:
                            if overlaps is None:
                                weight = 1
                            else:
                                weight = 1 - overlaps[self.internal2external_laneid[n]][i + 1]
                        elif any(
                            [
                                LaneRelation.neighbour_left in o or LaneRelation.neighbour_right in o
                                for o in self.g.get_edge_data(n, next_n)["label"]
                                + self.g.get_edge_data(next_n, n)["label"]
                            ]
                        ):
                            weight = 2
                        elif self.g.get_edge_data(n, next_n)["label"] in [
                            LaneRelation.predecessor,
                            LaneRelation.successor,
                        ] or self.g.get_edge_data(n, next_n)["label"] in [
                            LaneRelation.predecessor,
                            LaneRelation.successor,
                        ]:
                            weight = 2
                        else:
                            weight = 3
                    except Exception:
                        weight = 4
                    g.add_edge((n, i), (next_n, i + 1), weight=weight)
        for n in nodes_per_time[0]:
            g.add_edge("start", (n, 0), weight=1)
        for n in nodes_per_time[-1]:
            g.add_edge((n, len(nodes_per_time) - 1), "end", weight=1)
        sp = nx.shortest_path(g, "start", "end", weight="weight")[1:-1]
        fixed_traveler_path = [self.internal2external_laneid[o[0]] for o in sp]

        overlaps = [traveler_lane_intersections[lid][i] for i, lid in enumerate(fixed_traveler_path)]
        assert not np.any(np.isnan(overlaps))
        return fixed_traveler_path

    def __repr__(self):
        return f"Locator({len(self.all_lanes)} lanes)<{id(self)}>"



import sys
from pathlib import Path

import betterosi
import numpy as np
from loguru import logger
from lxd_io import Dataset
from tqdm.auto import tqdm
import multiprocessing as mp
from ..map_odr import MapOdr
from ..recording import Recording
import polars as pl

__all__ = ["convert_lxd"]
logger.configure(handlers=[{"sink": sys.stdout, "level": "WARNING"}])

NANOS_PER_SEC = 1000000000  # 1 s

vct = betterosi.MovingObjectVehicleClassificationType
vehicles = {
    "Car": vct.TYPE_CAR,
    "car": vct.TYPE_CAR,
    "Truck": vct.TYPE_HEAVY_TRUCK,
    "truck_bus": vct.TYPE_HEAVY_TRUCK,
    "truck": vct.TYPE_HEAVY_TRUCK,
    "bicycle": vct.TYPE_BICYCLE,
    "van": vct.TYPE_DELIVERY_VAN,
}
pedestrians = {"pedestrian": betterosi.MovingObjectType.TYPE_PEDESTRIAN}


class DatasetConverter:
    def __init__(self, dataset_dir: Path) -> None:
        self._dataset = Dataset(dataset_dir)

    def get_recording_ids(self) -> list[int]:
        return self._dataset.recording_ids

    def get_recording_opendrive_path(self, recording_id: int) -> Path:
        return self._dataset.get_recording(recording_id).opendrive_map_file

    def rec2df(self, recording_id):
        rec = self._dataset.get_recording(recording_id)
        dt = 1 / rec.get_meta_data("frameRate")

        meta = rec._tracks_meta_data
        meta = meta.with_columns(
            pl.col("class")
            .map_elements(
                (
                    lambda x: betterosi.MovingObjectType.TYPE_VEHICLE
                    if x in vehicles
                    else betterosi.MovingObjectType.TYPE_PEDESTRIAN
                ),
                return_dtype=int,
            )
            .alias("type"),
            pl.col("class")
            .map_elements(
                (lambda x: betterosi.MovingObjectVehicleClassificationRole.ROLE_CIVIL if x in vehicles else -1),
                return_dtype=int,
            )
            .alias("role"),
            pl.col("class")
            .map_elements((lambda x: vehicles[x] if x in vehicles else -1), return_dtype=int)
            .alias("subtype"),
        )
        meta = meta.rename({"trackId": "idx"})

        tracks = rec._get_tracks_data()
        tracks = tracks.rename(
            {
                "xCenter": "x",
                "yCenter": "y",
                "xVelocity": "vel_x",
                "yVelocity": "vel_y",
                "xAcceleration": "acc_x",
                "yAcceleration": "acc_y",
                "trackId": "idx",
                # "width": "width",
                # "length": "length",
            }
        )
        tracks = tracks.join(meta.select(["idx", "role", "type", "subtype"]), on="idx", how="left")
        is_vehicle = pl.col("type") == betterosi.MovingObjectType.TYPE_VEHICLE
        is_bicycle = pl.col("subtype") == betterosi.MovingObjectVehicleClassificationType.TYPE_BICYCLE
        is_pedestrian = pl.col("type") == betterosi.MovingObjectType.TYPE_PEDESTRIAN
        tracks = tracks.with_columns(
            [pl.lit(0.0).alias(k) for k in ["acc_z", "z", "vel_z", "roll", "pitch"]]
            + [
                (((pl.col("heading") + np.pi) % (2 * np.pi)) - np.pi).alias("yaw"),
                (pl.col("frame") * dt * NANOS_PER_SEC).cast(pl.Int64).alias("total_nanos"),
                pl.when(is_vehicle & is_bicycle)
                .then(0.8)
                .when(is_pedestrian)
                .then(0.5)
                .otherwise(pl.col("width"))
                .alias("width"),
                pl.when(is_vehicle & is_bicycle)
                .then(2.0)
                .when(is_pedestrian)
                .then(0.5)
                .otherwise(pl.col("length"))
                .alias("length"),
                pl.when(is_vehicle & is_bicycle).then(1.9).when(is_pedestrian).then(1.8).otherwise(2.0).alias("height"),
            ]
        )

        return tracks


def convert_recording(args):
    converter, recording_id, out_filename = args
    tracks = converter.rec2df(recording_id)
    xodr_path = converter.get_recording_opendrive_path(recording_id)
    rec = Recording(df=tracks, map=MapOdr.from_file(xodr_path), validate=False)
    rec.to_mcap(out_filename)


def convert_lxd(dataset_dir: Path, outpath: Path, n_workers=1):
    if n_workers == -1:
        n_workers = mp.cpu_count() - 1
    outpath = Path(outpath)
    dataset_dir = Path(dataset_dir)
    outpath.mkdir(exist_ok=True)
    converter = DatasetConverter(dataset_dir)
    args_list = [
        [converter, recording_id, outpath / f"{str(recording_id).zfill(2)}_tracks.mcap"]
        for recording_id in converter.get_recording_ids()
    ]

    if n_workers > 1:
        with mp.Pool(n_workers, maxtasksperchild=1) as pool:
            work_iterator = pool.imap(convert_recording, args_list, chunksize=1)
            list(tqdm(work_iterator, total=len(args_list)))
    else:
        for args in tqdm(args_list):
            convert_recording(args)



from dataclasses import dataclass, field
from typing import Any

import betterosi
import numpy as np
import shapely
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon as PltPolygon


@dataclass
class ProjectionOffset:
    x: float
    y: float
    z: float = 0.0
    yaw: float = 0.0


@dataclass(repr=False)
class LaneBoundary:
    _map: "Map" = field(init=False)
    idx: Any
    type: betterosi.LaneBoundaryClassificationType
    polyline: shapely.LineString
    # reference: Any = field(init=False, default=None)

    def plot(self, ax: plt.Axes):
        ax.plot(*np.array(self.polyline.coords).T, color="gray", alpha=0.1)

    def get_osi(self) -> betterosi.LaneBoundary:
        raise NotImplementedError()

    @classmethod
    def create(cls, *args, **kwargs):
        raise NotImplementedError()


@dataclass(repr=False)
class LaneBoundaryOsi(LaneBoundary):
    _osi: betterosi.LaneBoundary

    @classmethod
    def create(cls, lane_boundary: betterosi.LaneBoundary):
        return cls(
            idx=lane_boundary.id.value,
            polyline=shapely.LineString([(p.position.x, p.position.y) for p in lane_boundary.boundary_line]),
            type=betterosi.LaneBoundaryClassificationType(lane_boundary.classification.type),
            _osi=lane_boundary,
        )

    def get_osi(self) -> betterosi.LaneBoundary:
        return self._osi


@dataclass(repr=False)
class LaneBase:
    _map: "Map" = field(init=False)
    idx: Any
    centerline: np.ndarray
    type: betterosi.LaneClassificationType
    subtype: betterosi.LaneClassificationSubtype
    successor_ids: list[Any]
    predecessor_ids: list[Any]
    source_reference: Any = field(init=False)


@dataclass(repr=False)
class Lane(LaneBase):
    right_boundary_id: Any
    left_boundary_id: Any
    polygon: shapely.Polygon = field(init=False)
    left_boundary: LaneBoundary = field(init=False)
    right_boundary: LaneBoundary = field(init=False)
    # source_reference: Any = None


@dataclass(repr=False)
class LaneOsiCenterline(LaneBase):
    _osi: betterosi.Lane

    @staticmethod
    def _get_centerline(lane: betterosi.Lane):
        cl = np.array([(p.x, p.y) for p in lane.classification.centerline])
        if not lane.classification.centerline_is_driving_direction:
            cl = np.flip(cl, axis=0)
        return cl

    @classmethod
    def create(cls, lane: betterosi.Lane):
        return cls(
            _osi=lane,
            idx=lane.id.value,
            centerline=cls._get_centerline(lane),
            type=betterosi.LaneClassificationType(lane.classification.type),
            subtype=betterosi.LaneClassificationSubtype(lane.classification.subtype),
            successor_ids=[],
            predecessor_ids=[],
        )

    def plot(self, ax: plt.Axes):
        c = "black" if not self.type == betterosi.LaneClassificationType.TYPE_INTERSECTION else "green"
        ax.plot(*np.array(self.centerline).T, color=c, alpha=0.3, zorder=-10)


@dataclass(repr=False)
class LaneOsi(Lane, LaneOsiCenterline):
    right_boundary_ids: list[int]
    left_boundary_ids: list[int]
    free_boundary_ids: list[int]

    @classmethod
    def create(cls, lane: betterosi.Lane):
        return cls(
            _osi=lane,
            idx=int(lane.id.value),
            centerline=cls._get_centerline(lane),
            type=betterosi.LaneClassificationType(lane.classification.type),
            subtype=betterosi.LaneClassificationSubtype(lane.classification.subtype),
            successor_ids=[
                p.successor_lane_id.value for p in lane.classification.lane_pairing if p.successor_lane_id is not None
            ],
            predecessor_ids=[
                p.antecessor_lane_id.value for p in lane.classification.lane_pairing if p.antecessor_lane_id is not None
            ],
            right_boundary_ids=[idx.value for idx in lane.classification.right_lane_boundary_id if idx is not None],
            left_boundary_ids=[idx.value for idx in lane.classification.left_lane_boundary_id if idx is not None],
            right_boundary_id=[idx.value for idx in lane.classification.right_lane_boundary_id if idx is not None][0],
            left_boundary_id=[idx.value for idx in lane.classification.left_lane_boundary_id if idx is not None][0],
            free_boundary_ids=[idx.value for idx in lane.classification.free_lane_boundary_id if idx is not None],
        )

    def set_boundaries(self):
        self.left_boundary = self._map.lane_boundaries[self.left_boundary_ids[0]]
        self.right_boundary = self._map.lane_boundaries[self.right_boundary_ids[0]]

        # for omega
        self.oriented_borders = self._get_oriented_borders()
        self.start_points = np.array([b.interpolate(0, normalized=True) for b in self.oriented_borders])
        self.end_points = np.array([b.interpolate(1, normalized=True) for b in self.oriented_borders])
        return self

    def set_polygon(self):
        self.polygon = shapely.Polygon(
            np.concatenate(
                [
                    np.array(self.left_boundary.polyline.coords),
                    np.flip(np.array(self.right_boundary.polyline.coords), axis=0),
                ]
            )
        )
        if not self.polygon.is_simple:
            self.polygon = shapely.convex_hull(self.polygon)
        # TODO: fix or warning

    def plot(self, ax: plt.Axes):
        c = "green" if not self.type == betterosi.LaneClassificationType.TYPE_INTERSECTION else "black"
        ax.plot(*np.array(self.centerline).T, color=c, alpha=0.5)
        ax.add_patch(PltPolygon(self.polygon.exterior.coords, fc="blue", alpha=0.2, ec="black"))

    # for ase_engine/omega_prime
    def _get_oriented_borders(self):
        center_start = shapely.LineString(self.centerline).interpolate(0, normalized=True)
        left = self.left_boundary.polyline
        invert_left = left.project(center_start, normalized=True) > 0.5
        if invert_left:
            left = shapely.reverse(left)
        right = self.right_boundary.polyline
        invert_right = right.project(center_start, normalized=True) > 0.5
        if invert_right:
            right = shapely.reverse(right)
        return left, right


@dataclass(repr=False)
class Map:
    lane_boundaries: dict[int, LaneBoundary]
    lanes: dict[int:Lane]

    def plot(self, ax: plt.Axes):
        for l in self.lanes.values():
            l.plot(ax)
        for b in self.lane_boundaries.values():
            b.plot(ax)

    @classmethod
    def create(cls, *args, **kwargs):
        raise NotImplementedError()

    def __post_init__(self):
        self.setup_lanes_and_boundaries()

    def setup_lanes_and_boundaries(self):
        raise NotImplementedError()


@dataclass(repr=False)
class MapOsi(Map):
    _osi: betterosi.GroundTruth

    @classmethod
    def create(cls, gt: betterosi.GroundTruth):
        if len(gt.lane_boundary) == 0:
            raise RuntimeError("Empty Map")
        return cls(
            _osi=gt,
            lane_boundaries={b.id.value: LaneBoundaryOsi.create(b) for b in gt.lane_boundary},
            lanes={l.id.value: LaneOsi.create(l) for l in gt.lane if len(l.classification.right_lane_boundary_id) > 0},
        )

    def __post_init__(self):
        self.setup_lanes_and_boundaries()

    def setup_lanes_and_boundaries(self):
        for b in self.lane_boundaries.values():
            b._map = self
        for l in self.lanes.values():
            l._map = self
            l.set_boundaries()
            l.set_polygon()


@dataclass(repr=False)
class MapOsiCenterline(Map):
    _osi: betterosi.GroundTruth
    lanes: dict[int, LaneOsiCenterline]

    @classmethod
    def create(cls, gt: betterosi.GroundTruth):
        if len(gt.lane) == 0:
            raise RuntimeError("No Map")
        return cls(
            _osi=gt,
            lanes={l.id.value: LaneOsiCenterline.create(l) for l in gt.lane},
            lane_boundaries={},
        )

    def setup_lanes_and_boundaries(self):
        for l in self.lanes.values():
            l._map = self



from pathlib import Path
from typing import Annotated

import polars as pl
import typer
from pathlib import Path
from mcap_protobuf.decoder import DecoderFactory
from mcap.reader import make_reader
import omega_prime

app = typer.Typer(pretty_exceptions_show_locals=False)

app.registered_commands += omega_prime.converters.app.registered_commands


@app.command(help="Convert from ASAM OSI GroundTruth trace.")
def from_osi(
    input: Annotated[
        Path, typer.Argument(exists=True, dir_okay=False, help="Path to ASAM OSI trace file (either `.osi` or `.mcap`)")
    ],
    output: Annotated[Path, typer.Argument(exists=False, dir_okay=False, help="Desired filename of omega file")],
    odr: Annotated[
        Path | None, typer.Option(exists=True, dir_okay=False, help="Path to ASAM OpenDRIVE xml to use as map")
    ] = None,
    validate: bool = True,
):
    r = omega_prime.Recording.from_file(input, xodr_path=odr, validate=validate)
    r.to_mcap(output)


@app.command(help="Convert from csv table according to schema")
def from_csv(
    input: Annotated[
        Path,
        typer.Argument(exists=True, dir_okay=False, help="Path to csv according to omega moving object csv schema"),
    ],
    output: Annotated[Path, typer.Argument(exists=False, dir_okay=False, help="Desired filename of omega file")],
    odr: Annotated[
        Path | None, typer.Option(exists=True, dir_okay=False, help="Path to ASAM OpenDRIVE xml to use as map")
    ] = None,
    validate: bool = True,
):
    df = pl.read_csv(input)
    r = omega_prime.Recording(df, validate=validate)
    if odr is not None:
        r.map = omega_prime.MapOdr.from_file(odr)
    r.to_mcap(output)


@app.command(help="Check an omega-prime file for specification conformance.")
def validate(
    input: Annotated[Path, typer.Argument(help="Path to omega file to validate", exists=True, dir_okay=False)],
):
    omega_prime.Recording.from_file(input, validate=True, parse_map=True)
    print(f"File {input} is valid.")


@app.command(help="Extracts the ASAM OpenDRIVE file from the omega-prime file.")
def to_odr(
    input: Annotated[Path, typer.Argument(exists=True, dir_okay=False, help="Path to the omega-prime mcap file.")],
    output: Annotated[
        Path | None,
        typer.Argument(
            help="Where to store the ASAM OpenDRIVE file. If None or directory, stored filename will be used."
        ),
    ] = None,
):
    if Path(input).suffix == ".mcap":
        with Path(input).open("rb") as f:
            reader = make_reader(f, decoder_factories=[DecoderFactory()])
            gm = next(reader.iter_decoded_messages(topics="/ground_truth_map"))
        map = omega_prime.MapOdr.create(odr_xml=gm[3].open_drive_xml, name="", step_size=0.01)
        map.to_file(output)
        return None

    r = omega_prime.Recording.from_file(input, validate=False)
    if isinstance(r.map, omega_prime.MapOdr):
        r.map.to_file(output)
    else:
        raise ValueError("The provided omega-prime file does not contain a map in ASAM OpenDRIVE format.")


@app.command(help="Converts MCAP or OSI omega-prime into a parquet file (beneficial for faster loading of large data).")
def to_parquet(
    input: Annotated[Path, typer.Argument(exists=True, dir_okay=False, help="Path to the omega-prime mcap file.")],
    output: Annotated[
        Path | None,
        typer.Argument(help="Where to store the parquet file. If None or directory, stored filename will be used."),
    ] = None,
):
    r = omega_prime.Recording.from_file(input, validate=False, parse_map=False)
    if output is None:
        output = Path(input).parent / f"{Path(input).stem}.parquet"
    r.to_parquet(output)


if __name__ == "__main__":
    app()



import logging
from dataclasses import dataclass
from omega_prime.map import Map, Lane, LaneBoundary
import shapely
import numpy as np
from matplotlib.patches import Polygon as PltPolygon
import matplotlib.pyplot as plt
from betterosi import LaneClassificationType, LaneClassificationSubtype, LaneBoundaryClassificationType
from pyxodr.road_objects.network import RoadNetwork as PyxodrRoadNetwork
from pyxodr.road_objects.road import Road as PyxodrRoad
from pyxodr.road_objects.lane import Lane as PyxodrLane
from pathlib import Path
from lxml import etree
import pyproj
from .map import ProjectionOffset
from pathlib import Path
import betterosi
from betterosi import MapAsamOpenDrive
from collections import namedtuple

logger = logging.getLogger(__name__)


class RoadNetwork(PyxodrRoadNetwork):
    def __init__(
        self,
        xml_string: str,
        resolution: float = 0.1,
    ):
        self.root = etree.fromstring(xml_string.encode("utf-8"))
        self.tree = etree.ElementTree(self.root)
        self.resolution = resolution
        self.ignored_lane_types = set([])
        self.road_ids_to_object = {}


LANE_BOUNDARY_TYPE_MAP = {
    "none": LaneBoundaryClassificationType.TYPE_NO_LINE,
    "solid": LaneBoundaryClassificationType.TYPE_SOLID_LINE,
    "broken": LaneBoundaryClassificationType.TYPE_DASHED_LINE,
    "botts dots": LaneBoundaryClassificationType.TYPE_BOTTS_DOTS,
    "broken broken": LaneBoundaryClassificationType.TYPE_DASHED_LINE,
    "broken solid": LaneBoundaryClassificationType.TYPE_DASHED_LINE,
    "solid broken": LaneBoundaryClassificationType.TYPE_SOLID_LINE,
    "solid solid": LaneBoundaryClassificationType.TYPE_SOLID_LINE,
    "curb": LaneBoundaryClassificationType.TYPE_CURB,
    "edge": LaneBoundaryClassificationType.TYPE_ROAD_EDGE,
    "grass": LaneBoundaryClassificationType.TYPE_GRASS_EDGE,
    "custom": LaneBoundaryClassificationType.TYPE_OTHER,
}

LANE_SUBTYPE_MAP = {
    "driving": LaneClassificationSubtype.SUBTYPE_NORMAL,
    "shoulder": LaneClassificationSubtype.SUBTYPE_SHOULDER,
    "sidewalk": LaneClassificationSubtype.SUBTYPE_SIDEWALK,
    "walking": LaneClassificationSubtype.SUBTYPE_SIDEWALK,
    "parking": LaneClassificationSubtype.SUBTYPE_PARKING,
    "biking": LaneClassificationSubtype.SUBTYPE_BIKING,
    "stop": LaneClassificationSubtype.SUBTYPE_STOP,
    "restricted": LaneClassificationSubtype.SUBTYPE_RESTRICTED,
    "border": LaneClassificationSubtype.SUBTYPE_BORDER,
    "curb": LaneClassificationSubtype.SUBTYPE_BORDER,
    "exit": LaneClassificationSubtype.SUBTYPE_EXIT,
    "mwyexit": LaneClassificationSubtype.SUBTYPE_EXIT,
    "entry": LaneClassificationSubtype.SUBTYPE_ENTRY,
    "mwyentry": LaneClassificationSubtype.SUBTYPE_ENTRY,
    "onramp": LaneClassificationSubtype.SUBTYPE_ONRAMP,
    "offramp": LaneClassificationSubtype.SUBTYPE_OFFRAMP,
    "connectingramp": LaneClassificationSubtype.SUBTYPE_CONNECTINGRAMP,
}

OTHER_LANE_TYPES = {
    "bus",
    "taxi",
    "hov",
    "median",
    "sliplane",
    "shared",
    "tram",
    "rail",
    "none",
    "roadworks",
    "bidirectional",
    "special1",
    "special2",
    "special3",
}


# Define the named tuple
XodrIdx = namedtuple("XodrIdx", ["road_id", "lane_id", "section_idx", "side"])


def make_xodr_idx(road_id, lane_id, section_idx, side=None):
    return XodrIdx(road_id=road_id, lane_id=lane_id, section_idx=section_idx, side=side)


@dataclass(repr=False)
class MapOdr(Map):
    odr_xml: str
    name: str
    step_size: float = 0.01
    _xodr_map: PyxodrRoadNetwork | None = None
    proj_string: str | None = None
    proj_offset: ProjectionOffset | None = None
    projection: pyproj.CRS | None = None

    @property
    def xodr_map(self):
        if self._xodr_map is None:
            self.parse()
        return self._xodr_map

    @classmethod
    def from_file(
        cls,
        filename,
        topic="ground_truth_map",
        parse: bool = False,
        is_odr_xml: bool = False,
        is_mcap: bool = False,
        step_size=0.01,
    ):
        if Path(filename).suffix in [".xodr", ".odr"] or is_odr_xml:
            with open(filename) as f:
                odr_xml = f.read()
            return cls.create(odr_xml=odr_xml, name=Path(filename).stem, step_size=step_size, parse=parse)
        if Path(filename).suffix in [".mcap"] or is_mcap:
            map = next(iter(betterosi.read(filename, mcap_topics=[topic], osi_message_type=betterosi.MapAsamOpenDrive)))
            return cls.create(
                odr_xml=map.open_drive_xml_content, name=map.map_reference, step_size=step_size, parse=parse
            )

    @property
    def lanes(self):
        if self._lanes is None:
            self.parse()
        return self._lanes

    @lanes.setter
    def lanes(self, val):
        self._lanes = val

    @property
    def lane_boundaries(self):
        if self._lane_boundaries is None:
            self.parse()
        return self._lane_boundaries

    @lane_boundaries.setter
    def lane_boundaries(self, val):
        self._lane_boundaries = val

    @classmethod
    def create(cls, odr_xml, name, step_size=0.01, parse: bool = False):
        self = cls(odr_xml=odr_xml, name=name, step_size=step_size, lanes={}, lane_boundaries={})
        self._lane_boundaries = None
        self._lanes = None
        if parse:
            self.parse()
        return self

    def parse(self):
        rn = RoadNetwork(self.odr_xml, resolution=self.step_size)

        lane_boundaries = {}
        lanes = {}

        # Extract projection information from XML tree
        proj_string = None
        proj_offset = None
        projection = None

        # Get the header element from XML
        header = rn.tree.find(".//header")
        if header is not None:
            # Get geoReference if it exists
            geo_ref = header.find("geoReference")
            if geo_ref is not None and geo_ref.text:
                proj_string = geo_ref.text.strip()
                try:
                    projection = pyproj.CRS.from_proj4(proj_string)
                except pyproj.exceptions.CRSError as e:
                    logger.warning(f"Failed to parse projection string: {e}")

            # Get offset if it exists
            offset = header.find("offset")
            if offset is not None:
                try:
                    proj_offset = ProjectionOffset(
                        x=float(offset.get("x", "0")),
                        y=float(offset.get("y", "0")),
                        z=float(offset.get("z", "0")),
                        yaw=float(offset.get("hdg", "0")),
                    )
                except (ValueError, TypeError) as e:
                    logger.warning(f"Failed to parse offset: {e}")

        for road in rn.get_roads():
            lane_idx = 0
            for lane_section_idx, lane_section in enumerate(road.lane_sections):
                for lane in lane_section.lanes:
                    boundary_line = getattr(lane, "boundary_line", None)
                    if boundary_line is None or not len(boundary_line):
                        logger.warning(
                            f"Skipping road {road.id} / lane_section {lane_section_idx} / lane {lane.id}: missing boundary_line"
                        )
                        continue

                    try:
                        left_boundary = LaneBoundaryXodr.create(
                            lane, road.id, lane.id, lane_section_idx, "left", lane_idx=lane_idx
                        )
                        right_boundary = LaneBoundaryXodr.create(
                            lane, road.id, lane.id, lane_section_idx, "right", lane_idx=lane_idx
                        )
                    except Exception as e:
                        logger.error(
                            f"Failed to create boundaries for road {road.id} / lane_section {lane_section_idx} / lane {lane.id}: {e}"
                        )
                        continue

                    lane_boundaries[left_boundary.xodr_idx] = left_boundary
                    lane_boundaries[right_boundary.xodr_idx] = right_boundary

                    try:
                        lane_obj = LaneXodr.create(lane, road, lane_section_idx, lane_idx)
                        lanes[lane_obj.xodr_idx] = lane_obj
                    except Exception as e:
                        logger.error(
                            f"Failed to create lane object for road {road.id} / lane_section {lane_section_idx} / lane {lane.id}: {e}"
                        )

                    lane_idx += 1

        self._xodr_map = rn
        self.lane_boundaries = lane_boundaries
        self.lanes = lanes
        self.proj_string = proj_string
        self.proj_offset = proj_offset
        self.projection = projection
        for lane in self.lanes.values():
            lane._map = self
            lane.set_boundaries()
            lane.set_polygon()
        return self

    def setup_lanes_and_boundaries(self):
        for b in self.lane_boundaries.values():
            b._map = self

    def to_file(self, filename: str | Path):
        """Export the current MapOdr to a .xodr file."""
        if isinstance(filename, str):
            filename = Path(filename)

        if filename.is_dir():
            filename = filename / f"{self.name}.xodr"

        if filename.suffix == "":
            filename = filename.with_suffix(".xodr")

        with open(filename, "w") as f:
            f.write(self.odr_xml)

    def to_osi(self):
        return MapAsamOpenDrive(map_reference=self.name, open_drive_xml_content=self.odr_xml)

    def to_hdf(self, filename: Path | str):
        import tables

        # Ensure we have a path to the HDF5 file
        with tables.open_file(filename, mode="a") as h5file:
            # Ensure we have a /map group in the HDF5 file
            try:
                gmap = h5file.get_node("/map")
            except tables.NoSuchNodeError:
                gmap = h5file.create_group("/", "map")

            atom = tables.StringAtom(itemsize=len(self.odr_xml))
            ds = h5file.create_carray(gmap, "odr_xml", atom, shape=(1,))
            ds[0] = self.odr_xml.encode("utf-8")

    def from_hdf(cls, filename: Path | str):
        import tables

        with tables.open_file(filename, mode="r") as h5file:
            odr_xml = h5file.get_node("/map/odr_xml")[0].decode()
        return cls.create(odr_xml, str(filename))


@dataclass(repr=False)
class LaneBoundaryXodr(LaneBoundary):
    _xodr: PyxodrLane
    xodr_idx: str

    @classmethod
    def create(
        cls,
        boundary: PyxodrLane,
        road_id: str,
        lane_id: str,
        lane_section_idx,
        side: str,
        type: str = None,
        lane_idx: int = None,
    ):
        if side == "left":
            polyline = shapely.LineString(boundary.boundary_line)
        elif side == "right":
            polyline = shapely.LineString(boundary.lane_reference_line)
        else:
            raise ValueError(f"Invalid side '{side}'. Expected 'left' or 'right'.")

        if type is None and hasattr(boundary, "lane_xml"):
            type = cls._extract_lane_boundary_type_from_xml(boundary, side)

        lane_boundary_type = cls._determine_lane_boundary_type(type)

        return cls(
            idx=(int(road_id), lane_idx),
            xodr_idx=make_xodr_idx(road_id=road_id, section_idx=lane_section_idx, lane_id=lane_id, side=side),
            type=lane_boundary_type,
            polyline=np.array(polyline.coords),
            _xodr=boundary,
        )

    def plot(self, ax: plt.Axes):
        ax.plot(*np.array(self.polyline[:, :2]).T, color="gray", alpha=0.1)
        ax.set_aspect(1)

    @staticmethod
    def _extract_lane_boundary_type_from_xml(lane, side: str) -> str:
        if not hasattr(lane, "lane_xml"):
            return None
        road_marks = lane.lane_xml.findall("roadMark")
        try:
            if side == "right" and len(road_marks) > 0:
                if float(road_marks[0].get("sOffset", "inf")) == 0.0:
                    return road_marks[0].get("type")
            elif side == "left" and len(road_marks) > 1:
                return road_marks[1].get("type")
        except Exception:
            return None
        return None

    @staticmethod
    def _determine_lane_boundary_type(boundary_type_str: str) -> LaneBoundaryClassificationType:
        return LANE_BOUNDARY_TYPE_MAP.get(
            (boundary_type_str or "unknown").strip().lower(), LaneBoundaryClassificationType.TYPE_UNKNOWN
        )


@dataclass(repr=False)
class LaneXodr(Lane):
    _xodr: PyxodrLane
    xodr_idx: str

    @classmethod
    def create(cls, lane: PyxodrLane, road: PyxodrRoad, lane_section_idx: int, lane_idx: int = None):
        centre_line = getattr(lane, "centre_line", None)
        if centre_line is None or not len(centre_line):
            raise ValueError(f"Lane {lane.id} has no centre_line")

        centerline_2d = np.array([(p[0], p[1]) for p in lane.centre_line])
        lane_type, lane_subtype = cls._determine_lane_type_and_subtype(lane, road)

        return cls(
            _xodr=lane,
            idx=(int(road.id), lane_idx),
            xodr_idx=make_xodr_idx(road_id=road.id, section_idx=lane_section_idx, lane_id=lane.id, side=None),
            centerline=centerline_2d,
            type=lane_type,
            subtype=lane_subtype,
            successor_ids=[(int(road.id), lane_section_idx + 1, successor) for successor in lane.successor_ids],
            predecessor_ids=[(int(road.id), lane_section_idx - 1, predecessor) for predecessor in lane.predecessor_ids],
            right_boundary_id=make_xodr_idx(
                road_id=road.id, section_idx=lane_section_idx, side="right", lane_id=lane.id
            ),
            left_boundary_id=make_xodr_idx(road_id=road.id, section_idx=lane_section_idx, side="left", lane_id=lane.id),
        )

    @staticmethod
    def _determine_lane_type_and_subtype(lane: PyxodrLane, road: PyxodrRoad):
        lane_type = (
            LaneClassificationType.TYPE_UNKNOWN
            if road.road_xml.get("junction") == "-1"
            else LaneClassificationType.TYPE_INTERSECTION
        )
        lane_type_str = (getattr(lane, "type", "unknown") or "unknown").lower()

        if lane_type_str in LANE_SUBTYPE_MAP:
            lane_subtype = LANE_SUBTYPE_MAP[lane_type_str]
        elif lane_type_str in OTHER_LANE_TYPES:
            lane_subtype = LaneClassificationSubtype.SUBTYPE_OTHER
        else:
            lane_subtype = LaneClassificationSubtype.SUBTYPE_UNKNOWN

        return lane_type, lane_subtype

    def set_boundaries(self):
        self.left_boundary = self._map.lane_boundaries[self.left_boundary_id]
        self.right_boundary = self._map.lane_boundaries[self.right_boundary_id]
        return self

    def set_polygon(self):
        coords = np.concatenate(
            [
                self.left_boundary.polyline,
                np.flip(self.right_boundary.polyline, axis=0),
            ]
        )
        polygon = shapely.Polygon(coords)
        if not polygon.is_valid:
            polygon = polygon.buffer(0)
        self.polygon = polygon
        return self

    def plot(self, ax: plt.Axes):
        c = "green" if self.type != LaneClassificationType.TYPE_INTERSECTION else "black"
        ax.plot(*np.array(self.centerline).T, color=c, alpha=0.5)
        # try:
        if isinstance(self.polygon, shapely.geometry.MultiPolygon):
            ps = self.polygon.geoms
        else:
            ps = [self.polygon]
        for p in ps:
            ax.add_patch(PltPolygon(p.exterior.coords, fc="blue", alpha=0.2, ec="black"))
